//CUDA GPU FIR ver.7
// ring buffer in
//  -- separate --
// ring Buffer Output


// includes, system
#include <math.h>
#include <stdio.h>
#include <windows.h>

// includes, project
#include <cutil.h>
#include <cuda.h>

// Include, PortAudio
#include "pa_asio.h"
#include "pa_converters.h"

static void Float32_To_Int24(
    void *destinationBuffer, signed int destinationStride,
    void *sourceBuffer, signed int sourceStride,
    unsigned int count);
static void Float32_To_Int16(
    void *destinationBuffer, signed int destinationStride,
    void *sourceBuffer, signed int sourceStride,
    unsigned int count);

//PortAudio Parameters
#define SAMPLE_RATE  (44100)
#define FRAMES_PER_BUFFER (8192)

#define NUM_OUT_CHANNELS    (8)
#define PA_OUTPUT_SAMPLE_TYPE  paFloat32 //paInt24
#define OUTPUT_SAMPLE_SIZE (4)		//bytes
//#define PA_OUTPUT_SAMPLE_TYPE  paInt24 //paInt24
//#define OUTPUT_SAMPLE_SIZE (3)		//bytes

#define NUM_IN_CHANNELS    (2)
#define PA_INPUT_SAMPLE_TYPE  paInt16
#define INPUT_SAMPLE_SIZE (2)		//bytes

#define PA_ENABLE_DEBUG_OUTPUT

//CUDA Execution Parameter
#define CUDABLOCKS  16
#define THREAD_NUM  512		//executed thread count per block, do not change. shared memory is common in the block.

//FIR Processing parameters
#define TAPS		(8192)
#define WAY			4
#define LR			2
#define PACKET		(8192*1)	//GPU calculation length, 1+1 is too wasting but it works.

// includes, CUDA kernels
#include <WaveXGen08_kernel.cu>

////////////////////////////////////////////////////////////////////////////////
// declaration, forward

//FIR Related
void  ReadFIRCoeff(void);					//Read desired Frequency
void  ReadConfig(char * ConfigFile);		//Read Configuration
int  ProcessGPU(void * userData, PaStream *);	//Prototype Routine
int	CheckGPU(void);							//Check CUDA Available
float  sigmoid(float input);				//Peak Compression

//Ring Buffer between CUDA - PortAudio
void InitRingBuffer(void * userData);		
void FreeRingBuffer(void * userData);

//PortAudio Related
void  SetUpStreamParameter(PaStreamParameters * inputParameters, PaStreamParameters * outputParameters, int devicenum);
static int paCallback( const void *, void * ,
                            unsigned long ,
                            const PaStreamCallbackTimeInfo* ,
                            PaStreamCallbackFlags ,
                            void * );
//static int paCallback_OUT( const void *, void * ,
//                            unsigned long ,
//                            const PaStreamCallbackTimeInfo* ,
//                            PaStreamCallbackFlags ,
//                            void * );
static void PrintSupportedStandardSampleRates(
        const PaStreamParameters *inputParameters,
        const PaStreamParameters *outputParameters );

//Global Variables and structs
char	ConfigFile[256];

int		GPUDevice;			//Index of CUDA device
int		ASIODevice;			//ASIO device index
int inputChannelSelectors[2];	//ASIO input channel to use. all 1.
int outputChannelSelectors[8];	//ASIO output channel to use. all 1.
PaAsioStreamInfo ASIOInfoIN, ASIOInfoOUT;	//ASIO specific information
PaStreamParameters inputParameters, outputParameters;

struct FIRParams{
	char	CoeffFile[256];		//Coeff File Name
	float	FIRCoeff[TAPS];		//Coeff Value
};
struct FIRParams FIR[WAY];

CRITICAL_SECTION	cs;
HANDLE				hEvent;

typedef struct
{
	int	StopPlaying;	//flag should be related to signal from external GUI
	PaStream *pastream;	//to query stream information
	//PaStream *stream_OUT;	//to query stream information

	int numOutBytes;	//Output Block Size
	int numInBytes;		//Input Block Size

	//error handling
	unsigned long	LastStreamStatus;	

	//GPU FIR Parameter
	int PaddingLen;	//Padding Size
	int CalculationSize;	//Size of, actually to be calculated samples

	//Output Ring Buffer
	int PAOutCursor;	//0 to 7, Cursor TO READ
	int GPUOutCursor;	//0 to 7, Cursor TO WRITE
	int OutFilledFlag[8];	//filled(produced) = 1, blank(consumed) = 0
	unsigned char * pOutRing[8];	//Buffer contents pointer
	unsigned char * outDummy;		//dummy out buffer to test

	//Input Ring Buffer
	int PAInCursor;
	int GPUInCursor;
	int InFilledFlag[4];
	unsigned char * pInRing[4];
	unsigned char * inDummy;		//dummy in buffer to test

} paUserData;

////////////////////////////////////////////////////////////////////////////////
// Program main
////////////////////////////////////////////////////////////////////////////////
int main( int argc, char** argv) 
{
	//Check Arguments Count
	if (argc < 2)
	{
		printf("requires 1 arguments.\n");
		printf("WaveX06.exe ConfigFilePath\n");
		return -1;
	}

	//read GPU FIR configuration
	strncpy_s(ConfigFile,256, argv[1],strlen(argv[1]));
	ReadConfig(ConfigFile);
	ReadFIRCoeff();

    PaStream *pastream = NULL;
    //PaStream *stream_OUT = NULL;
    PaError err;
    paUserData data;
	data.pastream = pastream;
	//data.stream_OUT = stream_OUT;
	//Block size for output buffer.
	data.numOutBytes = FRAMES_PER_BUFFER * NUM_OUT_CHANNELS * OUTPUT_SAMPLE_SIZE;
	data.numInBytes = FRAMES_PER_BUFFER * NUM_IN_CHANNELS * INPUT_SAMPLE_SIZE;
	InitRingBuffer(&data);

    err = Pa_Initialize();
    if( err != paNoError ) goto error;

	SetUpStreamParameter(&inputParameters, &outputParameters, ASIODevice);

	//PrintSupportedStandardSampleRates(&inputParameters, &outputParameters);

	//confirm
	//printf("%d\n", inputParameters->hostApiSpecificStreamInfo.hostAPIType);
	//printf("%d\n", outputParameters->hostApiSpecificStreamInfo.hostAPIType);

    // format check
	if( Pa_IsFormatSupported( &inputParameters, 0, SAMPLE_RATE ) == paFormatIsSupported  )
	{
		printf( "Pa_IsFormatSupported reports input device will support %d channels, rate %d.\n", NUM_IN_CHANNELS, SAMPLE_RATE );
	}
	else
	{
		printf( "Pa_IsFormatSupported reports input device will NOT support %d channels, rate %d.\n", NUM_IN_CHANNELS, SAMPLE_RATE );
		goto error;
	}
	if( Pa_IsFormatSupported( 0, &outputParameters, SAMPLE_RATE ) == paFormatIsSupported  )
	{
		printf( "Pa_IsFormatSupported reports output device will support %d channels, rate %d.\n", NUM_OUT_CHANNELS, SAMPLE_RATE );
	}
	else
	{
		printf( "Pa_IsFormatSupported reports output device will NOT support %d channels, rate %d.\n", NUM_OUT_CHANNELS , SAMPLE_RATE);
		goto error;
	}

	InitializeCriticalSection(&cs);
	hEvent = CreateEvent(NULL, true, false, "PA_GPU_EVENT");
	//Update Process Priority
	SetPriorityClass(GetCurrentProcess(), ABOVE_NORMAL_PRIORITY_CLASS);

	err = CheckGPU();
    if( err != 0 ) goto error;

    err = Pa_OpenStream(
              &pastream,
              &inputParameters,
              &outputParameters,						
              SAMPLE_RATE,
              FRAMES_PER_BUFFER,
              paClipOff,
              paCallback,
			  &data ); /* userData */
	data.pastream = pastream;
    if( err != paNoError ) goto error;
	//start stream before process, let stream wait.
    err = Pa_StartStream( pastream );
    if( err != paNoError ) goto error;
    printf("PA stream started.\n");

 //   err = Pa_OpenStream(
 //             &stream_OUT,
 //             NULL,						//No input stream
 //             &outputParameters,
 //             SAMPLE_RATE,
 //             FRAMES_PER_BUFFER,
 //             paClipOff,
 //             paCallback_OUT,
	//		  &data ); /* userData */
	//data.stream_OUT = stream_OUT;

 //   if( err != paNoError ) goto error;

 //   err = Pa_StartStream( stream_OUT );
 //   if( err != paNoError ) goto error;
 //   printf("OUT stream started.\n");

	//Process
	while(data.StopPlaying == 0)	//should be some signal
	{
		//Test: when stream stop
		ProcessGPU(&data, pastream);
		//printf("\ndata.LastStreamStatus : %d .\n", data.LastStreamStatus);
		//printf("\nRestart Stream.\n");
		//err = Pa_AbortStream( stream_IN );
		//err = Pa_AbortStream( stream_OUT );
		//Pa_Sleep(200);
		//InitRingBuffer(&data);
		//err = Pa_StartStream( stream_IN );
		//err = Pa_StartStream( stream_OUT );
	}
  //  while( 1 )
  //  {
		//Pa_Sleep( 1000 );
  //     //err = Pa_WriteStream( stream, sampleBlock, FRAMES_PER_BUFFER );
  //     //if( err ) goto xrun;
  //     //err = Pa_ReadStream( stream, sampleBlock, FRAMES_PER_BUFFER );
  //     //if( err ) goto xrun;
  //  }
	DeleteCriticalSection(&cs);
	CloseHandle(hEvent);

	Pa_Sleep(50);
	FreeRingBuffer(&data);

	printf("%s\n",Pa_GetErrorText(err));

    err = Pa_StopStream( pastream );
    //err = Pa_StopStream( stream_OUT );
    if( err != paNoError ) goto error;

	Pa_Sleep(50);
	Pa_CloseStream( pastream );
	//Pa_CloseStream( stream_OUT );
	Pa_Sleep(50);
    Pa_Terminate();
    CUT_EXIT(argc, argv);

	return 0;
	
error:
    if( pastream ) {
       Pa_AbortStream( pastream );
       Pa_CloseStream( pastream );
    }
    //if( stream_OUT ) {
    //   Pa_AbortStream( stream_OUT );
    //   Pa_CloseStream( stream_OUT );
    //}
    Pa_Terminate();
    fprintf( stderr, "An error occured while using the portaudio stream\n" );
    fprintf( stderr, "Error number: %d\n", err );
    fprintf( stderr, "Error message: %s\n", Pa_GetErrorText( err ) );

	const PaHostErrorInfo * pahosterror;
	pahosterror = Pa_GetLastHostErrorInfo();
    fprintf( stderr, "Host Error type: %d\n", pahosterror->hostApiType);
    fprintf( stderr, "Host Error number: %d\n", pahosterror->errorCode );
    fprintf( stderr, "Host Error message: %s\n", pahosterror->errorText);


    CUT_EXIT(argc, argv);

	return -1;

}

void  SetUpStreamParameter(PaStreamParameters * inputParameters, PaStreamParameters * outputParameters, int devicenum)
{
    int indeviceIndex, outdeviceIndex;

	//Select Input Device which supports ASIO API
	indeviceIndex = Pa_GetHostApiInfo( Pa_HostApiTypeIdToHostApiIndex( paASIO ) )->defaultInputDevice;

	printf( "using device id %d (%s)\n", indeviceIndex, Pa_GetDeviceInfo(indeviceIndex)->name );

    inputParameters->device = indeviceIndex; /* DirectSound input device */

    printf( "input device # %d.\n", inputParameters->device );
	printf( "device maxInput : %d\n", Pa_GetDeviceInfo(inputParameters->device)->maxInputChannels);
	printf( "device maxOutput : %d\n", Pa_GetDeviceInfo(inputParameters->device)->maxOutputChannels);
    printf( "input LL: %g s\n", Pa_GetDeviceInfo( inputParameters->device )->defaultLowInputLatency );
    printf( "input HL: %g s\n", Pa_GetDeviceInfo( inputParameters->device )->defaultHighInputLatency );
    inputParameters->channelCount = NUM_IN_CHANNELS;
    inputParameters->sampleFormat = PA_INPUT_SAMPLE_TYPE;
    inputParameters->suggestedLatency = Pa_GetDeviceInfo( inputParameters->device )->defaultHighInputLatency;

	ASIOInfoIN.size = sizeof(PaAsioStreamInfo);
	ASIOInfoIN.hostApiType = paASIO;
	ASIOInfoIN.version = 1;
	//ASIOInfoIN.flags = 0;
	ASIOInfoIN.flags = paAsioUseChannelSelectors;	//FLAGS!
	inputChannelSelectors[0] = 0;
	inputChannelSelectors[1] = 1;
	//ASIOInfoIN.channelSelectors = NULL;
	ASIOInfoIN.channelSelectors = inputChannelSelectors;	//CHANNEL SELECTOR!
    inputParameters->hostApiSpecificStreamInfo = &ASIOInfoIN;

	//Select Output Device which supports ASIO API
	outdeviceIndex = Pa_GetHostApiInfo( Pa_HostApiTypeIdToHostApiIndex( paASIO ) )->defaultOutputDevice;

	printf( "using device id %d (%s)\n", outdeviceIndex, Pa_GetDeviceInfo(outdeviceIndex)->name );

    outputParameters->device = outdeviceIndex; /* DirectSound output device */

    printf( "Output device # %d.\n", outputParameters->device );
    printf( "Output LL: %g s\n", Pa_GetDeviceInfo( outputParameters->device )->defaultLowOutputLatency );
    printf( "Output HL: %g s\n", Pa_GetDeviceInfo( outputParameters->device )->defaultHighOutputLatency );
    outputParameters->channelCount = NUM_OUT_CHANNELS;
    outputParameters->sampleFormat = PA_OUTPUT_SAMPLE_TYPE;
    outputParameters->suggestedLatency = Pa_GetDeviceInfo( outputParameters->device )->defaultHighOutputLatency;

	ASIOInfoOUT.size = sizeof(PaAsioStreamInfo);
	ASIOInfoOUT.hostApiType = paASIO;
	ASIOInfoOUT.version = 1;
	ASIOInfoOUT.flags = paAsioUseChannelSelectors;
	outputChannelSelectors[0] = 0;
	outputChannelSelectors[1] = 1;
	outputChannelSelectors[2] = 2;
	outputChannelSelectors[3] = 3;
	outputChannelSelectors[4] = 4;
	outputChannelSelectors[5] = 5;
	outputChannelSelectors[6] = 6;
	outputChannelSelectors[7] = 7;
	ASIOInfoOUT.channelSelectors = outputChannelSelectors;
    outputParameters->hostApiSpecificStreamInfo = &ASIOInfoOUT;


}

void  ReadConfig(char* FileName)
{
	for (int i = 0; i < WAY; i++)
	{
		memset(FIR[i].CoeffFile, 0, sizeof(FIR[i].CoeffFile));
		for (int j = 0; j < TAPS; j++)
		{
			FIR[i].FIRCoeff[j] = 0.0;
		}
	}
	//apply global
	GPUDevice = GetPrivateProfileInt("Global_Params", "GPUIndex", 0, FileName);
	ASIODevice = GetPrivateProfileInt("Global_Params", "OutputDevice", -1, FileName);

	GetPrivateProfileString("CH12", "COEFF", "", FIR[0].CoeffFile, 256, FileName);
	GetPrivateProfileString("CH34", "COEFF", "", FIR[1].CoeffFile, 256, FileName);
	GetPrivateProfileString("CH56", "COEFF", "", FIR[2].CoeffFile, 256, FileName);
	GetPrivateProfileString("CH78", "COEFF", "", FIR[3].CoeffFile, 256, FileName);

}

void  ReadFIRCoeff(void)
{
	FILE * f;
	char buf[64];

	for (int j = 0; j < WAY; j++)
	{
		fopen_s(&f, FIR[j].CoeffFile,"r");
		if (f  != NULL)
		{
			int i = 0;
			double tmp = 0.0;
			while((!feof(f)) && (i < TAPS))
			{
				fgets(buf, sizeof(buf), f);
				sscanf_s(buf, "%lf", &tmp);
				FIR[j].FIRCoeff[i] = (float)tmp;
				i++;
			}
			fclose(f);
		}
	}
}

void InitRingBuffer(void * userData)
{
    paUserData *data = (paUserData*)userData;

	data->LastStreamStatus = 0;
	data->StopPlaying = 0;

	data->PAOutCursor = 0;
	data->GPUOutCursor = 0;
	data->PAInCursor = 0;
	data->GPUInCursor = 0;

	for (int i = 0; i < 4; i++)
	{
		data->InFilledFlag[i] = 0;
		data->pInRing[i] = (unsigned char *)malloc(data->numInBytes);
		VirtualLock(data->pInRing[i], data->numInBytes);
		memset(data->pInRing[i], 0, data->numInBytes);
	}
	for (int i = 0; i < 8; i++)
	{
		data->OutFilledFlag[i] = 0;
		data->pOutRing[i] = (unsigned char *)malloc(data->numOutBytes);
		VirtualLock(data->pOutRing[i], data->numOutBytes);
		memset(data->pOutRing[i], 0, data->numOutBytes);
	}
}

void FreeRingBuffer(void * userData)
{
    paUserData *data = (paUserData*)userData;

	//Free dummy buffer
	free(data->outDummy);
	free(data->inDummy);

	//Free Ring Buffer
	for (int i = 0; i < 4; i++)
	{
		free(data->pOutRing[i]);
		free(data->pInRing[i]);
	}
}


int CheckGPU(void)
{
	//CUT_DEVICE_INIT(argc, argv);
    int deviceCount;                                                         
	CUDA_SAFE_CALL_NO_SYNC(cudaGetDeviceCount(&deviceCount));                
    if (deviceCount == 0) {                                                  
        fprintf(stderr, "cutil error: no devices supporting CUDA.\n");       
        return -10;                                                  
    }

	//using device
    cudaDeviceProp deviceProp;                                               
    CUDA_SAFE_CALL_NO_SYNC(cudaGetDeviceProperties(&deviceProp, GPUDevice));       
    if (deviceProp.major < 1) {                                              
        fprintf(stderr, "cutil error: device does not support CUDA.\n");     
        return -10;                                                  
    }

    printf("\nDevice %d: \"%s\"\n", GPUDevice, deviceProp.name);
    printf("  Major revision number:                         %d\n",
           deviceProp.major);
    printf("  Minor revision number:                         %d\n",
           deviceProp.minor);
    printf("  Total amount of global memory:                 %u bytes\n",
           deviceProp.totalGlobalMem);
    printf("  Number of multiprocessors:                     %d\n",
           deviceProp.multiProcessorCount);
    printf("  Number of cores:                               %d\n",
           8 * deviceProp.multiProcessorCount);
    printf("  Total amount of constant memory:               %u bytes\n",
           deviceProp.totalConstMem); 
    printf("  Total amount of shared memory per block:       %u bytes\n",
           deviceProp.sharedMemPerBlock);
    printf("  Total number of registers available per block: %d\n",
           deviceProp.regsPerBlock);
    printf("  Warp size:                                     %d\n",
           deviceProp.warpSize);
    printf("  Maximum number of threads per block:           %d\n",
           deviceProp.maxThreadsPerBlock);
    printf("  Maximum sizes of each dimension of a block:    %d x %d x %d\n",
           deviceProp.maxThreadsDim[0],
           deviceProp.maxThreadsDim[1],
           deviceProp.maxThreadsDim[2]);
    printf("  Maximum sizes of each dimension of a grid:     %d x %d x %d\n",
           deviceProp.maxGridSize[0],
           deviceProp.maxGridSize[1],
           deviceProp.maxGridSize[2]);
    printf("  Maximum memory pitch:                          %u bytes\n",
           deviceProp.memPitch);
    printf("  Texture alignment:                             %u bytes\n",
           deviceProp.textureAlignment);
    printf("  Clock rate:                                    %.2f GHz\n",
           deviceProp.clockRate * 1e-6f);
    printf("  Concurrent copy and execution:                 %s\n",
           deviceProp.deviceOverlap ? "Yes" : "No");
	
	CUDA_SAFE_CALL(cudaSetDevice(GPUDevice));                                      
	return 0;
}

int  ProcessGPU( void * userData, PaStream * stream_IN )
{

	//Variables
	//h_idata			: memory, Host, to put into GPU.
	//d_coeff			: memory, GPU, to store coefficient.size is WAY*TAPS
	//d_idata			: memory, GPU, to store input wave data. length needs packetlen + coeff len.
	//d_odata			: memory, GPU, result data stored. size is equal to input packet size. (not including coeffs tail size)
	//h_odata[8]		: memory, Host, to store result. there are total 8 (4Way * L,R) buffer. size is same as packet.
	//fWave_Out_Data	: memory, Host, to store Packet result. type = float. PACKET*sizeof(float)*LR*WAY 
	//FIR[j].FIRCoeff[m]: Channel FIR Coeff.

    paUserData *data = (paUserData*)userData;
	int StreamStatus = 1;

	//Prepare Local Buffer
	unsigned char * sWave_In_Data;
	sWave_In_Data = (unsigned char *)malloc(data->numInBytes * 2);	//Current and Next
	VirtualLock(sWave_In_Data, data->numInBytes * 2);
	memset(sWave_In_Data, 0, data->numInBytes * 2);

	//Prepare Host and Device Memory

	//Host Data to Go GPU
	float * h_idata;	
	cudaError_t cret;
	cret = cudaMallocHost((void**)&h_idata, sizeof(float)*(PACKET + TAPS));
	if (cret != CUDA_SUCCESS)
	{
        printf("Can not allocate %d bytes host memory.\n", sizeof(float)*(PACKET + TAPS));
        return -1;
	}
	
	//GPU device input memory
	float * d_idata;
	CUDA_SAFE_CALL(cudaMalloc((void**) &d_idata, sizeof(float)*(PACKET + TAPS)) );

	//GPU device memory for result
	float * d_odata;
	CUDA_SAFE_CALL(cudaMalloc((void**) &d_odata, sizeof(float)*PACKET) );

	//host output (from GPU) memory, to consolidate results for one sampling frame
	float * h_odata[8];	//for WAY way * (L&R)
	for (int i = 0 ; i < 8; i++)
	{
		cret = cudaMallocHost((void**)&(h_odata[i]), sizeof(float)*PACKET);
		if (cret != CUDA_SUCCESS)
		{
			printf("Can not allocate %d bytes host memory.\n", sizeof(float)*PACKET);
			return -1;
		}
	}

	//allocate host WAVE format result buffer
	float * fWave_Out_Data;
	cret = cudaMallocHost((void**)&fWave_Out_Data, PACKET*sizeof(float)*LR*WAY);
	if (cret != CUDA_SUCCESS)
	{
        printf("Can not allocate %d bytes host memory.\n", PACKET*sizeof(float)*LR*WAY);
        return -1;
	}

    // setup parallel execution parameters
    dim3  grid( CUDABLOCKS, 1, 1);
    dim3  threads( THREAD_NUM, 1, 1);

	int FrameMSec = (FRAMES_PER_BUFFER*1000)/SAMPLE_RATE;
	long EventStatus;

	while(StreamStatus > 0)
	{

		//for starting only. not smart.
		//if ( ( Pa_IsStreamActive(data->pastream) == 0) && (data->GPUOutCursor > 3))
		//{
		//	Pa_StartStream( data->pastream );
		//	printf("OUT stream started.\n");
		//}

		//Wait for Stream CallBack Event, up to 3 frames
		EventStatus = WaitForSingleObject(hEvent, FrameMSec * 3);
		Pa_Sleep(10);	//give 10msec to loopback function finishes,
		ResetEvent(hEvent);

		//ABANDONED 0x0080 , TIMEOUT 0x0102	
		if (EventStatus > 0)
		{
			StreamStatus = Pa_IsStreamActive(stream_IN);
			printf("\nFIR processing Event Error. ret = 0x%lx, IsStreamActive = %d\n", EventStatus, StreamStatus);
			goto FINISH;
		}

		// Check Current Cursor, Next Cursor
		int waitflag = 0;	//sleep while loop

		EnterCriticalSection(&cs);
		int NextCursor = data->GPUInCursor + 1;
		if (NextCursor > 3) NextCursor = 0;
		// GPUed FIR requires additional frame to run parallel.
		if ((data->InFilledFlag[data->GPUInCursor] == 0) || (data->InFilledFlag[NextCursor] == 0))
		{
			waitflag = 1;
		}
		else
		{
			//Copy Input Buffer to Local
			memcpy(sWave_In_Data                   , data->pInRing[ data->GPUInCursor ], data->numInBytes);
			memcpy(sWave_In_Data + data->numInBytes, data->pInRing[ NextCursor        ], data->numInBytes);
			memset(data->pInRing[ data->GPUInCursor ], 0, data->numInBytes); 
			
			//Channel Delay will be implemented here.
			//(retain previous - to move on top - offset copy next)

			//Input Buffer was consumed.
			data->InFilledFlag[data->GPUInCursor] = 0;
			data->GPUInCursor++;
			//if end, back to top
			if (data->GPUInCursor > 3) data->GPUInCursor = 0;		
		}
        LeaveCriticalSection(&cs);

		// Wait for 2 input buffer filled.
		if (waitflag == 1)
		{
			Pa_Sleep(1);	//input buffer not ready. sleep a little.
			continue;		//while
		}

		//Loop for FIR[0], [1], [2], [3]: Needs to change Coeffs.
		for (int j = 0; j < WAY; j++)
		{
			//Loop for L, R: prepare PACKET words + TAP words to process.
			for (int k = 0; k < LR; k++)
			{
				unsigned char CDataL, CDataH;
				short LData;

				for (int m = 0; m < (PACKET + TAPS); m++)
				{
					CDataL = (unsigned char)( * (sWave_In_Data + m*LR*2 + k*LR    ));
					CDataH = (unsigned char)( * (sWave_In_Data + m*LR*2 + k*LR + 1));
					LData = (short)(CDataL + (CDataH<<8));

					h_idata[m] = ((float)(LData))/32768.0f;
				}

				//Copy Wave Data to Device
				CUDA_SAFE_CALL( cudaMemcpy(d_idata, h_idata,sizeof(float)*(PACKET + TAPS),cudaMemcpyHostToDevice));
				//copy coeffs to constant
				CUDA_SAFE_CALL( cudaMemcpyToSymbol(coeff_Kernel, FIR[j].FIRCoeff, sizeof(float)*TAPS) );

				//Call Kernel
				//CUDA_SAFE_CALL( cudaThreadSynchronize() );
				calcFIR<<<grid,threads>>>(d_idata, d_odata, PACKET);
				CUT_CHECK_ERROR("calcFIR failed");
				//CUDA_SAFE_CALL( cudaThreadSynchronize() );
				//Get processed data to Packet Output Buffer
				CUDA_SAFE_CALL( cudaMemcpy( (void *)(h_odata[j*LR + k]), d_odata, sizeof(float)*PACKET, cudaMemcpyDeviceToHost) );
			}//L, R Loop End		
		}//0,1,2,3 Loop End

		//8xPacket buffer ready
		//Copy to Wave Output Area
		float LValue;
		float RValue;
		for ( int j = 0; j < WAY; j++)
		{
			for (int i = 0; i < PACKET; i++)
			{
				LValue = (h_odata[j*LR  ][i]);
				RValue = (h_odata[j*LR+1][i]);
				//Channel dB offset will be implemented here.
				fWave_Out_Data[i*WAY*LR + j*LR    ] = sigmoid(LValue);
				fWave_Out_Data[i*WAY*LR + j*LR + 1] = sigmoid(RValue);
			}
		}
		
		//Output to Ring Buffer
		EnterCriticalSection(&cs);
		if (data->OutFilledFlag[data->GPUOutCursor] == 0)
		{
			//Buffer blank, or consumed. Copy FIR result.
			//Float32_To_Int24(data->pOutRing[data->GPUOutCursor], 1, fWave_Out_Data, 1, PACKET*WAY*LR*sizeof(float));
			//Float32_To_Int16(data->pOutRing[data->GPUOutCursor], 1, fWave_Out_Data, 1, PACKET*WAY*LR*sizeof(float));
			memcpy(data->pOutRing[data->GPUOutCursor], fWave_Out_Data, PACKET*WAY*LR*sizeof(float));
			//Buffer Filled.
			data->OutFilledFlag[data->GPUOutCursor] = 1;
			//Cursor +1
			data->GPUOutCursor++;
			//if end, back to top
			if (data->GPUOutCursor > 7) data->GPUOutCursor = 0;
		}
		LeaveCriticalSection(&cs);

		////Output to Ring Buffer
		//EnterCriticalSection(&cs);
		//if (data->OutFilledFlag[data->GPUOutCursor] == 0)
		//{
		//	//Buffer blank, or consumed. Copy FIR result.
		//	memcpy(data->pOutRing[data->GPUOutCursor], fWave_Out_Data, PACKET*WAY*LR*sizeof(float));
		//	//Buffer Filled.
		//	data->OutFilledFlag[data->GPUOutCursor] = 1;
		//	//Cursor +1
		//	data->GPUOutCursor++;
		//	//if end, back to top
		//	if (data->GPUOutCursor > 7) data->GPUOutCursor = 0;
		//}
		//LeaveCriticalSection(&cs);

//#ifdef _DEBUG
//		clock_t currenttime = clock();
//		printf("GPU  IFb%d%d%d%d PI%d GI%d OFb%d%d%d%d%d%d%d%d GO%d PO%d T: %8d\r", 
//			data->InFilledFlag[0],
//			data->InFilledFlag[1],
//			data->InFilledFlag[2],
//			data->InFilledFlag[3],
//			data->PAInCursor,
//			data->GPUInCursor,
//			data->OutFilledFlag[0],
//			data->OutFilledFlag[1],
//			data->OutFilledFlag[2],
//			data->OutFilledFlag[3],
//			data->OutFilledFlag[4],
//			data->OutFilledFlag[5],
//			data->OutFilledFlag[6],
//			data->OutFilledFlag[7],
//			data->PAOutCursor,
//			data->GPUOutCursor,
//			currenttime
//			);
//#endif
		//End Loop for Packet
	}

FINISH:
	printf("\n exit FIR processing.\n");
	//Free Memory
	cudaFreeHost(h_idata);
	cudaFree(d_idata);
	cudaFree(d_odata);
	for (int i = 0; i < 8; i++)
	{
		cudaFreeHost(h_odata[i]);
	}
	cudaFreeHost(fWave_Out_Data);
	free(sWave_In_Data);
	return 0;
}

float sigmoid(float input)
{
	double output;
	double sigmoid;
	sigmoid = 0.0;

	if (input > 0.5)
	{
		sigmoid = 1.0/(1.0 + exp(-4.0*(input - 0.5)));
		return (float)sigmoid;
	}
	if (input < -0.5)
	{
		sigmoid = 1.0/(1.0 + exp(-4.0*(input + 0.5))) - 1.0;
		return (float)sigmoid;
	}
	
	return (float)input;
}

//float sigmoid(float input)
//{
//	//sigmoid function, avoid clipping
//	if (input > 0.98)
//	{
//		input = 0.98f+ ( (1.0f/(1.0f + exp(-20.0f*(input-0.98f))) - 0.5f) * 0.1f);
//	}
//	if (input < -0.98)
//	{
//		input = -0.98f- (0.5f - (1.0f/(1.0f + exp(-20.0f*(input+0.98f))))) * 0.1f;
//	}
//	return input;
//}

static int paCallback( const void *inputBuffer, void *outputBuffer,
                            unsigned long framesPerBuffer,
                            const PaStreamCallbackTimeInfo* timeInfo,
                            PaStreamCallbackFlags statusFlags,
                            void *userData )
{
    paUserData *data = (paUserData*)userData;
    unsigned char *out = (unsigned char *)outputBuffer;
    unsigned char *in = (unsigned char *)inputBuffer;

    (void) timeInfo; /* Prevent unused variable warnings. */
    unsigned long StreamCallBackFlag = statusFlags;
	int returnValue = 0;

	int StreamValid = 0;
    //EnterCriticalSection(&cs);

	//Check for Stream input and status
	//short frame. do nothing.
	if (framesPerBuffer < FRAMES_PER_BUFFER) 
	{
		StreamValid = -1;
		StreamCallBackFlag = StreamCallBackFlag | 0x00000100;
		printf("\n short frame, framesPerBuffer=%d \n",framesPerBuffer);
		//Pa_Sleep(10);
	}
	//See PaStreamCallbackFlags portaudio.h
	if (StreamCallBackFlag > 0) 
	{
		StreamValid = -1;
		//overflow or underflow, input or output : continue.
		//when aborted, stream processing will stop. do not abort.
		printf("\n overflow or underflow, StreamCallBackFlag = 0x%x \n",StreamCallBackFlag);
		//Pa_Sleep(10);
		//return paContinue; there are no difference.
	}
	data->LastStreamStatus = StreamCallBackFlag;

	//Input Processing
	if (StreamValid == 0)
	{
		//Copy Audio Stream to GPU FIR Buffer
		memcpy(data->pInRing[data->PAInCursor], in, data->numInBytes);
	}
	else
	{
		//Run GPU with 00000000.
		memset(data->pInRing[data->PAInCursor], 0, data->numInBytes);
	}

	//Buffer Filled.
	data->InFilledFlag[data->PAInCursor] = 1;
	//Cursor +1
	data->PAInCursor++;
	//if end, back to top
	if (data->PAInCursor > 3) data->PAInCursor = 0;
	//Output Processing
	//Check Buffer Ready
	if (data->OutFilledFlag[data->PAOutCursor] == 1)
	{
		//Buffer ready. copy to output.
		memcpy(out, data->pOutRing[data->PAOutCursor],data->numOutBytes);
		//once used, cleanup.
		memset(data->pOutRing[data->PAOutCursor], 0, data->numOutBytes);
		//Buffer consumed.
		data->OutFilledFlag[data->PAOutCursor] = 0;
		//Cursor +1
		data->PAOutCursor++;
		//if end, back to top
		if (data->PAOutCursor > 7) data->PAOutCursor = 0;
	}
	else
	{
		//buffer is not ready.
		memset(out, 0, data->numOutBytes);
	}

#ifdef _DEBUG
	clock_t currenttime = clock();
	printf("PIN  IFb%d%d%d%d PI%d GI%d OFb%d%d%d%d%d%d%d%d GO%d PO%d T: %8d LEN %6d ST %d\r", 
			data->InFilledFlag[0],
			data->InFilledFlag[1],
			data->InFilledFlag[2],
			data->InFilledFlag[3],
			data->PAInCursor,
			data->GPUInCursor,
			data->OutFilledFlag[0],
			data->OutFilledFlag[1],
			data->OutFilledFlag[2],
			data->OutFilledFlag[3],
			data->OutFilledFlag[4],
			data->OutFilledFlag[5],
			data->OutFilledFlag[6],
			data->OutFilledFlag[7],
			data->PAOutCursor,
			data->GPUOutCursor,
			currenttime,
			framesPerBuffer,
			StreamCallBackFlag
			);
#endif

	//LeaveCriticalSection(&cs);

	SetEvent(hEvent);

	returnValue = paContinue;
	return returnValue;

}
static int paCallback_OUT( const void *inputBuffer, void *outputBuffer,
                            unsigned long framesPerBuffer,
                            const PaStreamCallbackTimeInfo* timeInfo,
                            PaStreamCallbackFlags statusFlags,
                            void *userData )
{
    paUserData *data = (paUserData*)userData;
    unsigned char *out = (unsigned char *)outputBuffer;
    (void) inputBuffer ;
    unsigned long StreamCallBackFlag = statusFlags;
    (void) timeInfo; /* Prevent unused variable warnings. */
	int returnValue = 0;

    //EnterCriticalSection(&cs);

	//Output Processing
	//Check Buffer Ready
	if (data->OutFilledFlag[data->PAOutCursor] == 1)
	{
		//Buffer ready. copy to output.
		memcpy(out, data->pOutRing[data->PAOutCursor],data->numOutBytes);
		//once used, cleanup.
		memset(data->pOutRing[data->PAOutCursor], 0, data->numOutBytes);
		//Buffer consumed.
		data->OutFilledFlag[data->PAOutCursor] = 0;
		//Cursor +1
		data->PAOutCursor++;
		//if end, back to top
		if (data->PAOutCursor > 7) data->PAOutCursor = 0;
	}
	else
	{
		//buffer is not ready.
		memset(out, 0, data->numOutBytes);
	}

#ifdef _DEBUG
	clock_t currenttime = clock();
	printf("POUT IFb%d%d%d%d PI%d GI%d OFb%d%d%d%d%d%d%d%d GO%d PO%d T: %8d LEN %6d ST %d\r", 
			data->InFilledFlag[0],
			data->InFilledFlag[1],
			data->InFilledFlag[2],
			data->InFilledFlag[3],
			data->PAInCursor,
			data->GPUInCursor,
			data->OutFilledFlag[0],
			data->OutFilledFlag[1],
			data->OutFilledFlag[2],
			data->OutFilledFlag[3],
			data->OutFilledFlag[4],
			data->OutFilledFlag[5],
			data->OutFilledFlag[6],
			data->OutFilledFlag[7],
			data->PAOutCursor,
			data->GPUOutCursor,
			currenttime,
			framesPerBuffer,
			StreamCallBackFlag
			);
#endif
	//LeaveCriticalSection(&cs);

	returnValue = paContinue;
	return returnValue;

}

static void Float32_To_Int24(
    void *destinationBuffer, signed int destinationStride,
    void *sourceBuffer, signed int sourceStride,
    unsigned int count)
{
    float *src = (float*)sourceBuffer;
    unsigned char *dest = (unsigned char*)destinationBuffer;
    signed int temp;
   
    while( count-- )
    {
        /* convert to 32 bit and drop the low 8 bits */
        double scaled = *src * 0x7FFFFFFF;
        temp = (signed int) scaled;
        dest[0] = (unsigned char)(temp >> 8);
        dest[1] = (unsigned char)(temp >> 16);
        dest[2] = (unsigned char)(temp >> 24);
        
        src += sourceStride;
        dest += destinationStride * 3;
    }
}

static void Float32_To_Int16(
    void *destinationBuffer, signed int destinationStride,
    void *sourceBuffer, signed int sourceStride,
    unsigned int count)
{
    float *src = (float*)sourceBuffer;
    signed short *dest =  (signed short *)destinationBuffer;

    while( count-- )
    {
        short samp = (short) (*src * (32767.0f));
        *dest = samp;

        src += sourceStride;
        dest += destinationStride;
    }
}

static void PrintSupportedStandardSampleRates(
        const PaStreamParameters *inputParameters,
        const PaStreamParameters *outputParameters )
{
    static double standardSampleRates[] = {
        8000.0, 9600.0, 11025.0, 12000.0, 16000.0, 22050.0, 24000.0, 32000.0,
        44100.0, 48000.0, 88200.0, 96000.0, 192000.0, -1 /* negative terminated  list */
    };
    int     i, printCount;
    PaError err;

    printCount = 0;
    for( i=0; standardSampleRates[i] > 0; i++ )
    {
        err = Pa_IsFormatSupported( inputParameters, outputParameters, standardSampleRates[i] );
        if( err == paFormatIsSupported )
        {
            if( printCount == 0 )
            {
                printf( "\t%8.2f", standardSampleRates[i] );
                printCount = 1;
            }
            else if( printCount == 4 )
            {
                printf( ",\n\t%8.2f", standardSampleRates[i] );
                printCount = 1;
            }
            else
            {
                printf( ", %8.2f", standardSampleRates[i] );
                ++printCount;
            }
        }
    }
    if( !printCount )
        printf( "None\n" );
    else
        printf( "\n" );
}